# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Gammapy integration and system tests.

This package can be used for tests that involved several
Gammapy sub-packages, or that don't fit anywhere else.
"""
